// (C) Copyright 2011 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software
// in object code form for any purpose and without fee is hereby
// granted, provided that the above copyright notice appears in
// all copies and that both that copyright notice and the limited
// warranty and restricted rights notice below appear in all
// supporting documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK,
// INC. DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL
// BE UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is
// subject to restrictions set forth in FAR 52.227-19 (Commercial
// Computer Software - Restricted Rights) and DFAR 252.227-7013(c)
// (1)(ii)(Rights in Technical Data and Computer Software), as
// applicable. 
//


using System;
using System.Collections.Generic;
using System.Linq;

using Autodesk.Revit.DB;
using Autodesk.Revit.DB.Architecture;
using Autodesk.Revit.UI;
using Autodesk.Revit.UI.Selection;
using Autodesk.Revit.ApplicationServices;
using Autodesk.Revit.Attributes;

/* Lab7 can copy the picked group to all selected rooms with the 
 * same position. It shows the usage of following basic APIs 
 * 
 *  Room.IsPointIn() to determine if a point is in a room.
 *  Element.BoundingBox to calculate the center of an element.
 *  Pick multiple elements with filter.
 *  
 * Before run this command, a group should  already be in the 
 * active document.
 */

[TransactionAttribute(TransactionMode.Manual)]
[RegenerationAttribute(RegenerationOption.Manual)]
public class Lab7SelectRoomsAndPlaceGroup : IExternalCommand
{
  public Result Execute(
    ExternalCommandData commandData,
    ref string message,
    ElementSet elements)
  {
    UIApplication uiApp = commandData.Application;    
    Document doc = uiApp.ActiveUIDocument.Document;

    try
    {
      Reference pickedRef = null;
      Selection sel = uiApp.ActiveUIDocument.Selection;

      // Pick a group
      GroupPickFilter selFilter = new GroupPickFilter();
      pickedRef = sel.PickObject(ObjectType.Element, selFilter, 
        "Please select a group");
      Element elem = doc.GetElement(pickedRef);
      Group group = (Group)elem;
        
      // Get the group's center point
      XYZ origin = GetElementCenter(group);

      // Get the room that the picked group is located in
      Room room = GetRoomOfGroup(doc, origin);

      // Get the room's center point
      XYZ sourceCenter = GetRoomCenter(room);

      // Ask the user to pick target rooms
      RoomPickFilter roomPickFilter = new RoomPickFilter();
      IList<Reference> rooms =
        sel.PickObjects(
          ObjectType.Element,
          roomPickFilter, 
          "Select target rooms for duplicate furniture group");

      // Place furniture in each of the rooms
      Transaction trans = new Transaction(doc);
      trans.Start("Lab");
      PlaceFurnitureInRooms(
        doc, rooms, sourceCenter,
        group.GroupType, origin);
      trans.Commit();
    }
    // If user right-clicks or presess Esc button,
    // handle the exception
    catch (Autodesk.Revit.Exceptions.OperationCanceledException)
    {
      return Result.Cancelled;
    }
    // Catch other errors
    catch (Exception ex)
    {
      message = ex.Message;
      return Result.Failed;
    }

    return Result.Succeeded;
  }

  /// <summary>
  /// Return the room in which the given point is located
  /// </summary>
  Room GetRoomOfGroup(Document doc, XYZ point)
  {
    FilteredElementCollector collector = 
      new FilteredElementCollector(doc);
    collector.OfCategory(BuiltInCategory.OST_Rooms);
    Room room = null;
    foreach (Element elem in collector)
    {
      room = elem as Room;
      if (room != null)
      {
        // Decide if this point is in the picked room                  
        if (room.IsPointInRoom(point))
        {
          break;
        }
      }
    }
    return room;
  }

  /// <summary>
  /// Return the center of an element based on its BoundingBox
  /// </summary>
  public XYZ GetElementCenter(Element elem)
  {
    BoundingBoxXYZ bounding = elem.get_BoundingBox(null);
    XYZ center = (bounding.Max + bounding.Min) * 0.5;
    return center;
  }

  /// <summary> 
  /// Return a room's center point coordinates. 
  /// Z value is equal to the bottom of the room
  /// </summary>
  public XYZ GetRoomCenter(Room room)
  {
    // Get the room center point.
    XYZ boundCenter = GetElementCenter(room);
    LocationPoint locPt = (LocationPoint)room.Location;
    XYZ roomCenter =
      new XYZ(boundCenter.X, boundCenter.Y, locPt.Point.Z);
    return roomCenter;
  }

  /// <summary>
  /// Copy the group to each of the provided rooms. The position
  /// at which the group should be placed is based on the target
  /// room's center point: it should have the same offset from
  /// this point as the original had from the center of its room
  /// </summary>
  public void PlaceFurnitureInRooms(
    Document doc,
    IList<Reference> rooms,
    XYZ sourceCenter,
    GroupType gt,
    XYZ groupOrigin)
  {
    XYZ offset = groupOrigin - sourceCenter;
    XYZ offsetXY = new XYZ(offset.X, offset.Y, 0);

    foreach (Reference r in rooms)
    {
      Room roomTarget = doc.GetElement(r) as Room;
      if (roomTarget != null)
      {
        XYZ roomCenter = GetRoomCenter(roomTarget);
        Group group =
          doc.Create.PlaceGroup(roomCenter + offsetXY, gt);
      }
    }
  }
}

/// <summary>
/// Filter to constrain picking to model groups
/// </summary>
public class GroupPickFilter : ISelectionFilter
{
  public bool AllowElement(Element e)
  {
    return (e.Category.Id.IntegerValue.Equals(
      (int)BuiltInCategory.OST_IOSModelGroups));
  }
  public bool AllowReference(Reference r, XYZ p)
  {
    return false;
  }
}

/// <summary>
/// Filter to constrain picking to rooms
/// </summary>
public class RoomPickFilter : ISelectionFilter
{
  public bool AllowElement(Element e)
  {
    return (e.Category.Id.IntegerValue.Equals(
      (int)BuiltInCategory.OST_Rooms));
  }

  public bool AllowReference(Reference r, XYZ p)
  {
    return false;
  }
}
